from typing import Callable

from requests import adapters
from urllib3 import util

from up42.http import config


def create(
    supply_settings: Callable[[], config.ResilienceSettings] = config.ResilienceSettings, include_post: bool = False
) -> adapters.HTTPAdapter:
    settings = supply_settings()
    allowed_methods = set(util.Retry.DEFAULT_ALLOWED_METHODS)
    if include_post:
        allowed_methods.add("POST")

    retries = util.Retry(
        total=settings.total,
        backoff_factor=settings.backoff_factor,
        status_forcelist=settings.statuses,
        allowed_methods=allowed_methods,
    )
    return adapters.HTTPAdapter(max_retries=retries)



import dataclasses
import datetime
import functools
import importlib.metadata
import json
import logging
import pathlib
import tarfile
import tempfile
import warnings
import zipfile
from typing import Any, Callable, List, Optional, Union, cast
from urllib import parse

import geojson  # type: ignore
import pystac_client
import requests
import tqdm

from up42 import host

TIMEOUT = 120  # seconds
CHUNK_SIZE = 1024


def get_filename(signed_url: str, default_filename: str) -> str:
    """
    Returns the filename from the signed URL
    """
    parsed_url = parse.urlparse(signed_url)
    extension = pathlib.Path(parsed_url.path).suffix
    try:
        file_name = parse.parse_qs(parsed_url.query)["response-content-disposition"][0].split("filename=")[1]
        return f"{file_name}{extension}"
    except (IndexError, KeyError):
        warnings.warn(
            f"Unable to extract filename from URL. Using default filename: {default_filename}",
            UserWarning,
        )
        return f"{default_filename}{extension}"


def get_logger(
    name: str,
    level=logging.INFO,
    verbose: bool = False,
):
    """
    Use level=logging.CRITICAL to disable temporarily.
    """
    inner_logger = logging.getLogger(name)
    inner_logger.setLevel(level)
    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(level)
    if verbose:
        log_format = "%(asctime)s - %(name)s - %(levelname)s - %(message)"
    else:
        # hide logger module & level, truncate log messages > 2000 characters (e.g. huge geometries)
        log_format = "%(asctime)s - %(message).2000s"
    formatter = logging.Formatter(log_format)
    ch.setFormatter(formatter)
    inner_logger.addHandler(ch)
    inner_logger.propagate = False
    return inner_logger


logger = get_logger(__name__)


def deprecation(
    replacement_name: Optional[str],
    version: str,
):
    """
    Decorator for custom deprecation warnings.

    Args:
        replacement_name: Name of the replacement function.
        version: The breaking package version
    """

    def actual_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            replace_with = f" Use `{replacement_name}` instead." if replacement_name else ""
            message = f"`{func.__name__}` is deprecated and will be removed in version {version}{replace_with}."
            warnings.warn(message, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        return wrapper

    return actual_decorator


def _unpack_tar_files(file_path: str, output_directory: Union[str, pathlib.Path]) -> List[pathlib.Path]:
    out_filepaths: List[pathlib.Path] = []
    if tarfile.is_tarfile(file_path):
        with tarfile.open(file_path) as tar_file:
            for tar_member in tar_file.getmembers():
                if tar_member.isfile():
                    if "output/" in tar_member.name:
                        tar_member.name = tar_member.name.split("output/")[1]
                    tar_file.extract(tar_member, output_directory)
                    out_filepaths.append(pathlib.Path(output_directory) / tar_member.name)
    return out_filepaths


def _unpack_zip_files(file_path: str, output_directory: Union[str, pathlib.Path]) -> List[pathlib.Path]:
    out_filepaths: List[pathlib.Path] = []
    if zipfile.is_zipfile(file_path):
        with zipfile.ZipFile(file_path) as zip_file:
            for zip_info in zip_file.infolist():
                if not zip_info.filename.endswith("/"):
                    if "output/" in zip_info.filename:
                        zip_info.filename = zip_info.filename.split("output/")[1]
                    zip_file.extract(zip_info, output_directory)
                    out_filepaths.append(pathlib.Path(output_directory) / zip_info.filename)
    return out_filepaths


def download_archive(
    download_url: str,
    output_directory: Union[str, pathlib.Path],
) -> List[str]:
    """
    General download function for results of storage assets, job & jobtask from cloud storage
    provider.

    Args:
        download_url: The signed gcs url to download.
        output_directory: The file output directory, defaults to the current working
            directory.
    """
    # Download
    with tempfile.NamedTemporaryFile(dir=output_directory) as dst:
        try:
            r = requests.get(download_url, stream=True, timeout=TIMEOUT)
            r.raise_for_status()
            for chunk in tqdm.tqdm(r.iter_content(chunk_size=CHUNK_SIZE)):
                if chunk:  # filter out keep-alive new chunks
                    dst.write(chunk)
            dst.flush()
        except requests.exceptions.HTTPError as err:
            error_message = f"Connection error, please try again! {err}"
            logger.debug(error_message)
            raise requests.exceptions.HTTPError(error_message)
        # Order results are zip, job results are tgz(tar.gzipped)
        out_filepaths = _unpack_tar_files(dst.name, output_directory) + _unpack_zip_files(dst.name, output_directory)

        if not out_filepaths:
            raise UnsupportedArchive("Downloaded file is not a TGZ/TAR or ZIP archive.")
        logger.info(
            "Download successful of %s files to output_directory %s",
            len(out_filepaths),
            output_directory,
        )
    return [str(p) for p in out_filepaths]


class UnsupportedArchive(ValueError):
    pass


@dataclasses.dataclass
class ImageFile:
    url: str
    file_name: str = "output"
    session: requests.Session = dataclasses.field(default=requests.session(), repr=False, compare=False)

    def download(self, output_directory: Union[str, pathlib.Path]) -> pathlib.Path:
        file_name = get_filename(self.url, default_filename=self.file_name)
        path = pathlib.Path().joinpath(output_directory, file_name)
        with open(path, "wb") as dst:
            try:
                r = self.session.get(self.url, stream=True, timeout=TIMEOUT)
                r.raise_for_status()
                for chunk in tqdm.tqdm(r.iter_content(chunk_size=CHUNK_SIZE)):
                    if chunk:  # filter out keep-alive new chunks
                        dst.write(chunk)
            except requests.exceptions.HTTPError as err:
                logger.debug("Connection error, please try again! %s", err)
                raise requests.exceptions.HTTPError(f"Connection error, please try again! {err}")

            logger.info("Successfully downloaded the file at %s", path)
            return path


def download_file(download_url: str, output_directory: Union[str, pathlib.Path]) -> List[str]:
    """
    General download function for assets, job and jobtasks from cloud storage
    provider.

    Args:
        download_url: The signed gcs url to download.
        output_directory: The file output directory, defaults to the current working
            directory.
    """
    image = ImageFile(download_url)
    return [str(image.download(output_directory))]


def format_time(date: Optional[Union[str, datetime.datetime]], set_end_of_day=False):
    """
    Formats date isostring to datetime string format

    Args:
        date: datetime object or isodatetime string e.g. "YYYY-MM-DD" or "YYYY-MM-DDTHH:MM:SS".
        set_end_of_day: Sets the date to end of day, as required for most image archive searches. Only applies for
            type date string without time, e.g. "YYYY-MM-DD", not explicit datetime object or time of day.
    """
    if isinstance(date, str):
        has_time_of_day = len(date) > 11
        date = datetime.datetime.fromisoformat(date)
        if not has_time_of_day and set_end_of_day:
            date = datetime.datetime.combine(date.date(), datetime.time(23, 59, 59, 999999))
    elif isinstance(date, datetime.datetime):
        pass
    else:
        raise ValueError("date needs to be of type datetime or isoformat date string!")

    return date.strftime("%Y-%m-%dT%H:%M:%SZ")


def validate_fc_up42_requirements(fc: Union[dict, geojson.FeatureCollection]):
    """
    Validate the feature collection if it fits UP42 geometry requirements.
    """
    geometry_error = "UP42 only accepts single geometries, the provided geometry {}."
    if len(fc["features"]) != 1:
        raise ValueError(geometry_error.format("contains multiple geometries"))

    fc_type = fc["features"][0]["geometry"]["type"]
    if fc_type != "Polygon":
        raise ValueError(geometry_error.format(f"is a {fc_type}"))


def get_up42_py_version():
    """Get the version of the up42-py package."""
    return importlib.metadata.version("up42-py")


def read_json(path_or_dict: Union[dict, str, pathlib.Path, None]) -> Optional[dict]:
    if path_or_dict and isinstance(path_or_dict, (str, pathlib.Path)):
        try:
            with open(path_or_dict, encoding="utf-8") as file:
                return json.load(file)
        except FileNotFoundError as ex:
            raise ValueError(f"File {path_or_dict} does not exist!") from ex
    return cast(Optional[dict], path_or_dict)


def stac_client(auth: requests.auth.AuthBase):
    # pystac client accepts both returning and non-returning request modifiers
    # requests.auth.AuthBase is a returning request modifier interface
    request_modifier = cast(Callable[[requests.Request], Optional[requests.Request]], auth)
    return pystac_client.Client.open(
        url=host.endpoint("/v2/assets/stac/"),
        request_modifier=request_modifier,
    )


@dataclasses.dataclass(frozen=True)
class SortingField:
    name: str
    ascending: bool = True

    @property
    def asc(self):
        return SortingField(name=self.name)

    @property
    def desc(self):
        return SortingField(name=self.name, ascending=False)

    def __str__(self):
        order = "asc" if self.ascending else "desc"
        return f"{self.name},{order}"


def paged_query(params: dict[str, Any], endpoint: str, session: requests.Session):
    params = {key: value for key, value in params.items() if value is not None}
    params["page"] = 0

    def get_pages():
        while True:
            response = session.get(host.endpoint(endpoint), params=params).json()
            yield response["content"]
            params["page"] += 1
            if params["page"] >= response["totalPages"]:
                break

    return (entry for page in get_pages() for entry in page)



import abc
import dataclasses
import datetime
import enum
from typing import ClassVar, Iterator, List, Optional, TypedDict, Union

import pystac
import requests
import tenacity as tnc

from up42 import base, host, utils

ISO_FORMAT_LENGTH = 23  # precision including milliseconds


@dataclasses.dataclass(frozen=True)
class ValidationError:
    message: str
    name: str


class JobStatus(enum.Enum):
    CREATED = "created"
    LICENSED = "licensed"
    UNLICENSED = "unlicensed"
    VALID = "valid"
    INVALID = "invalid"
    ACCEPTED = "accepted"
    REJECTED = "rejected"
    RUNNING = "running"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    CAPTURED = "captured"
    RELEASED = "released"


TERMINAL_STATUSES = [
    JobStatus.CAPTURED,
    JobStatus.RELEASED,
    JobStatus.INVALID,
    JobStatus.REJECTED,
    JobStatus.UNLICENSED,
]


class JobResults(TypedDict, total=False):
    collection: Optional[str]
    errors: Optional[List[dict]]


class JobMetadata(TypedDict):
    # pylint: disable=invalid-name
    processID: str
    jobID: str
    accountID: str
    workspaceID: Optional[str]
    definition: dict
    results: Optional[JobResults]
    creditConsumption: Optional[dict]
    status: str
    created: str
    started: Optional[str]
    finished: Optional[str]
    updated: str


class UnfinishedJob(Exception):
    """Job hasn't finished yet with success or failure"""


class JobSorting:
    process_id = utils.SortingField("processID")
    status = utils.SortingField("status", ascending=False)
    created = utils.SortingField("created", ascending=False)
    credits = utils.SortingField("creditConsumption.credits", ascending=False)


def _to_datetime(value: Optional[str]):
    return value and datetime.datetime.fromisoformat(value[:ISO_FORMAT_LENGTH])


@dataclasses.dataclass
class Job:
    session = base.Session()
    stac_client = base.StacClient()
    process_id: str
    id: str
    account_id: str
    workspace_id: Optional[str]
    definition: dict
    status: JobStatus
    created: datetime.datetime
    updated: datetime.datetime
    collection_url: Optional[str] = None
    errors: Optional[List[ValidationError]] = None
    credits: Optional[int] = None
    started: Optional[datetime.datetime] = None
    finished: Optional[datetime.datetime] = None

    @property
    def collection(self) -> Optional[pystac.Collection]:
        if self.collection_url is None:
            return None
        collection_id = self.collection_url.split("/")[-1]
        return self.stac_client.get_collection(collection_id)

    @staticmethod
    def from_metadata(metadata: JobMetadata) -> "Job":
        results: JobResults = metadata.get("results") or {}
        errors = results.get("errors") or []
        validation_errors = [ValidationError(**error) for error in errors]
        consumption = metadata.get("creditConsumption") or {}
        return Job(
            process_id=metadata["processID"],
            id=metadata["jobID"],
            account_id=metadata["accountID"],
            workspace_id=metadata["workspaceID"],
            collection_url=results.get("collection"),
            errors=validation_errors or None,
            credits=consumption.get("credits"),
            definition=metadata["definition"],
            status=JobStatus(metadata["status"]),
            created=_to_datetime(metadata["created"]),
            started=_to_datetime(metadata["started"]),
            finished=_to_datetime(metadata["finished"]),
            updated=_to_datetime(metadata["updated"]),
        )

    def track(self, *, wait: int = 60, retries: int = 60 * 24 * 3):
        @tnc.retry(
            stop=tnc.stop_after_attempt(retries),
            wait=tnc.wait_fixed(wait),
            retry=tnc.retry_if_exception_type(UnfinishedJob),
            reraise=True,
        )
        def update():
            job = Job.get(self.id)
            self.status = job.status
            self.updated = job.updated
            self.finished = job.finished
            self.started = job.started
            self.collection_url = job.collection_url
            self.errors = job.errors
            self.credits = job.credits
            if self.status not in TERMINAL_STATUSES:
                raise UnfinishedJob

        update()

    @classmethod
    def get(cls, job_id: str) -> "Job":
        url = host.endpoint(f"/v2/processing/jobs/{job_id}")
        metadata = cls.session.get(url).json()
        return cls.from_metadata(metadata)

    @classmethod
    def all(
        cls,
        process_id: Optional[List[str]] = None,
        workspace_id: Optional[str] = None,
        status: Optional[List[JobStatus]] = None,
        min_duration: Optional[int] = None,
        max_duration: Optional[int] = None,
        sort_by: Optional[utils.SortingField] = None,
        *,
        # used for performance tuning and testing only
        page_size: Optional[int] = None,
    ) -> Iterator["Job"]:
        query_params = {
            key: str(value)
            for key, value in {
                "workspaceId": workspace_id,
                "processId": ",".join(process_id) if process_id else None,
                "status": ",".join(entry.value for entry in status) if status else None,
                "minDuration": min_duration,
                "maxDuration": max_duration,
                "limit": page_size,
                "sort": sort_by,
            }.items()
            if value
        }

        def get_pages():
            page = cls.session.get(host.endpoint("/v2/processing/jobs"), params=query_params).json()
            while page:
                yield page["jobs"]
                next_page_url = next(
                    (link["href"] for link in page["links"] if link["rel"] == "next"),
                    None,
                )
                page = next_page_url and cls.session.get(host.endpoint(next_page_url)).json()

        for page in get_pages():
            for metadata in page:
                yield Job.from_metadata(metadata)


CostType = Union[int, float, "Cost"]


@dataclasses.dataclass(frozen=True)
class Cost:
    strategy: str
    credits: int
    size: Optional[int] = None
    unit: Optional[str] = None

    def __le__(self, other: CostType):
        if isinstance(other, Cost):
            return self.credits <= other.credits
        else:
            return self.credits <= other

    def __lt__(self, other: CostType):
        if isinstance(other, Cost):
            return self.credits < other.credits
        else:
            return self.credits < other

    def __ge__(self, other: CostType):
        return not self < other

    def __gt__(self, other: CostType):
        return not self <= other


class JobTemplate:
    session = base.Session()
    process_id: ClassVar[str]
    workspace_id: Union[str, base.WorkspaceId]
    errors: set[ValidationError] = set()
    process_description: dict = {}

    @property
    @abc.abstractmethod
    def inputs(self) -> dict:
        pass

    def __post_init__(self):
        self.__validate()
        if self.is_valid:
            self.__evaluate()

    def __validate(self):
        self.errors = self.__validate_process_exists() or self.__validate_eula() or self.__validate_inputs()

    def __validate_process_exists(self) -> set[ValidationError]:
        process_url = host.endpoint(f"/v2/processing/processes/{self.process_id}")
        try:
            self.process_description = self.session.get(process_url).json()
        except requests.HTTPError as err:
            if err.response.status_code == 404:
                return {
                    ValidationError(
                        name="ProcessNotFound",
                        message=f"The process {self.process_id} does not exist.",
                    )
                }
            else:
                raise err
        return set()

    def __validate_eula(self) -> set[ValidationError]:
        eula_id = next(
            parameter["value"][0]
            for parameter in self.process_description["additionalParameters"]["parameters"]
            if parameter["name"] == "eula-id"
        )
        eula_url = host.endpoint(f"/v2/eulas/{eula_id}")
        eula = self.session.get(eula_url).json()
        if not eula["isAccepted"]:
            return {
                ValidationError(
                    name="EulaNotAccepted",
                    message=f"EULA for the process {self.process_id} not accepted.",
                )
            }
        return set()

    def __validate_inputs(self) -> set[ValidationError]:
        url = host.endpoint(f"/v2/processing/processes/{self.process_id}/validation")
        try:
            _ = self.session.post(url, json={"inputs": self.inputs})
        except requests.HTTPError as err:
            if (status_code := err.response.status_code) in (400, 422):
                if status_code == 400:
                    return {
                        ValidationError(
                            name="InvalidSchema",
                            message=err.response.json()["schema-error"],
                        )
                    }
                if status_code == 422:
                    errors = err.response.json()["errors"]
                    return {ValidationError(**error) for error in errors}
            else:
                raise err
        return set()

    def __evaluate(self):
        url = host.endpoint(f"/v2/processing/processes/{self.process_id}/cost")
        payload = self.session.post(url, json={"inputs": self.inputs}).json()
        self.cost = Cost(
            strategy=payload["pricingStrategy"],
            credits=payload["totalCredits"],
            size=payload.get("totalSize"),
            unit=payload.get("unit"),
        )

    @property
    def is_valid(self) -> bool:
        return not self.errors

    def execute(self) -> Job:
        url = host.endpoint(f"/v2/processing/processes/{self.process_id}/execution")
        job_metadata = self.session.post(
            url, params={"workspaceId": self.workspace_id}, json={"inputs": self.inputs}
        ).json()
        return Job.from_metadata(job_metadata)


@dataclasses.dataclass
class SingleItemJobTemplate(JobTemplate):
    title: str
    item: pystac.Item

    @property
    def inputs(self) -> dict:
        return {"title": self.title, "item": self.item.get_self_href()}


@dataclasses.dataclass
class MultiItemJobTemplate(JobTemplate):
    title: str
    items: List[pystac.Item]

    @property
    def inputs(self) -> dict:
        return {
            "title": self.title,
            "items": [item.get_self_href() for item in self.items],
        }



import dataclasses
import enum
from typing import Any, Iterator, Literal, Optional, Union

import geojson  # type: ignore
import requests

from up42 import base, host, utils


class CollectionType(enum.Enum):
    ARCHIVE = "ARCHIVE"
    TASKING = "TASKING"


IntegrationValue = Literal[
    "ACCESS_APPROVAL_REQUIRED",
    "SAMPLE_DATA_AVAILABLE",
    "MANUAL_REQUEST_REQUIRED",
    "FEASIBILITY_STUDY_REQUIRED",
    "FEASIBILITY_STUDY_MAY_BE_REQUIRED",
    "QUOTATION_REQUIRED",
    "PRICE_ESTIMATION_AVAILABLE",
    "SEARCH_AVAILABLE",
    "THUMBNAIL_AVAILABLE",
    "QUICKLOOK_AVAILABLE",
]


@dataclasses.dataclass
class ResolutionValue:
    minimum: float
    maximum: Optional[float] = None


@dataclasses.dataclass
class CollectionMetadata:
    product_type: Optional[Literal["OPTICAL", "SAR", "ELEVATION"]]
    resolution_class: Optional[Literal["VERY_HIGH", "HIGH", "MEDIUM", "LOW"]]
    resolution_value: Optional[ResolutionValue]


BoundingBox = list[float]


@dataclasses.dataclass
class Scene:
    bbox: Optional[BoundingBox]
    geometry: Union[geojson.Polygon, geojson.MultiPolygon]
    id: str
    datetime: Optional[str]
    start_datetime: Optional[str]
    end_datetime: Optional[str]
    constellation: str
    collection: str
    cloud_coverage: Optional[float]
    resolution: Optional[float]
    delivery_time: Optional[Literal["MINUTES", "HOURS", "DAYS"]]
    producer: str
    quicklook: utils.ImageFile
    provider_properties: dict


class InvalidHost(ValueError):
    pass


class InvalidSearchRequest(ValueError):
    pass


@dataclasses.dataclass
class Provider:
    session = base.Session()
    name: str
    # TODO: switch to Optional[str] = None in 3.0.0
    title: str = ""
    # TODO: switch to Optional[str] = None in 3.0.0
    description: str = ""
    # TODO: switch Sequence[Literal["PRODUCER", "HOST"]] = ("HOST",) in 3.0.0
    roles: list[Literal["PRODUCER", "HOST"]] = dataclasses.field(default_factory=lambda: ["HOST"])

    @property
    def is_host(self):
        return "HOST" in self.roles

    def search(
        self,
        bbox: Optional[BoundingBox] = None,
        intersects: Optional[geojson.Polygon] = None,
        query: Optional[dict] = None,
        collections: Optional[list[str]] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
    ) -> Iterator[Scene]:
        if not self.is_host:
            raise InvalidHost("Provider does not host collections")
        datetime_str = None
        if start_date or end_date:
            start_datetime = utils.format_time(start_date) if start_date else ".."
            end_datetime = utils.format_time(end_date, set_end_of_day=True) if end_date else ".."
            datetime_str = f"{start_datetime}/{end_datetime}"

        payload = {
            key: value
            for key, value in {
                "bbox": bbox,
                "intersects": intersects,
                "datetime": datetime_str,
                "query": query,
                "collections": collections,
            }.items()
            if value
        }

        def get_pages():
            url = host.endpoint(f"/catalog/hosts/{self.name}/stac/search")
            while url:
                try:
                    page: dict = self.session.post(url, json=payload).json()
                    yield page["features"]
                    url = next(
                        (link["href"] for link in page["links"] if link["rel"] == "next"),
                        None,
                    )
                except requests.HTTPError as http_error:
                    if http_error.response.status_code == 422:
                        error = http_error.response.json()["error"]
                        raise InvalidSearchRequest(error["message"]) from http_error
                    raise http_error

        for page in get_pages():
            for feature in page:
                yield self._as_scene(feature)

    def _as_scene(self, feature: geojson.Feature) -> Scene:
        properties = feature["properties"]
        scene_id = properties["id"]
        return Scene(
            bbox=feature.get("bbox"),
            geometry=feature["geometry"],
            id=scene_id,
            constellation=properties["constellation"],
            collection=properties["collection"],
            producer=properties["producer"],
            datetime=properties.get("datetime"),
            start_datetime=properties.get("start_datetime"),
            end_datetime=properties.get("end_datetime"),
            cloud_coverage=properties.get("cloudCoverage"),
            resolution=properties.get("resolution"),
            delivery_time=properties.get("deliveryTime"),
            quicklook=utils.ImageFile(
                url=host.endpoint(f"/catalog/{self.name}/image/{scene_id}/quicklook"),
                file_name=f"quicklook_{scene_id}.jpg",
                session=self.session,
            ),
            provider_properties=properties["providerProperties"],
        )


@dataclasses.dataclass
class DataProduct:
    session = base.Session()
    name: str
    title: str
    description: str
    id: Optional[str]
    eula_id: Optional[str]

    @property
    def schema(self) -> Optional[dict]:
        if self.id:
            url = host.endpoint(f"/orders/schema/{self.id}")
            return self.session.get(url).json()
        else:
            return None


@dataclasses.dataclass
class Collection:
    name: str
    title: str
    description: str
    type: CollectionType
    integrations: list[IntegrationValue]
    providers: list[Provider]
    data_products: list[DataProduct]
    metadata: Optional[CollectionMetadata]


class CollectionSorting:
    name = utils.SortingField("name")
    title = utils.SortingField("title")
    description = utils.SortingField("description")
    type = utils.SortingField("type")


class ProductGlossary:
    session = base.Session()

    @classmethod
    def get_collections(
        cls,
        collection_type: Optional[CollectionType] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator[Collection]:
        query_params: dict[str, Any] = {"sort": sort_by} if sort_by else {}

        def get_pages():
            current_page = 0
            while True:
                query_params["page"] = current_page
                page = cls.session.get(host.endpoint("/v2/collections"), params=query_params).json()
                total_pages = page["totalPages"]
                yield page["content"]
                current_page += 1
                if current_page == total_pages:
                    break

        for page_content in get_pages():
            for collection in page_content:
                if collection_type is None or collection["type"] == collection_type.value:
                    metadata = collection.get("metadata")
                    yield Collection(
                        name=collection["name"],
                        title=collection["title"],
                        description=collection["description"],
                        type=CollectionType(collection["type"]),
                        integrations=collection["integrations"],
                        providers=[Provider(**provider) for provider in collection["providers"]],
                        data_products=[
                            DataProduct(
                                name=data_product["name"],
                                title=data_product["title"],
                                description=data_product["description"],
                                id=data_product.get("id"),
                                eula_id=data_product.get("eulaId"),
                            )
                            for data_product in collection["dataProducts"]
                        ],
                        metadata=metadata
                        and CollectionMetadata(
                            product_type=metadata.get("productType"),
                            resolution_class=metadata.get("resolutionClass"),
                            resolution_value=metadata.get("resolutionValue")
                            and ResolutionValue(**metadata.get("resolutionValue")),
                        ),
                    )



from typing import Callable

import requests

from up42 import utils
from up42.http import http_adapter

SCHEMAS = ["http", "https"]
REPOSITORY_URL = "https://github.com/up42/up42-py"
HttpAdapterFactory = Callable[[], requests.adapters.HTTPAdapter]


class StatusValidatingSession(requests.Session):
    def request(self, *args, **kwargs) -> requests.Response:
        raise_for_status = kwargs.pop("raise_for_status", True)
        response = super().request(*args, **kwargs)
        if raise_for_status:
            response.raise_for_status()
        return response


def create(
    auth: requests.auth.AuthBase,
    create_adapter: HttpAdapterFactory = http_adapter.create,
    version: str = utils.get_up42_py_version(),
) -> requests.Session:
    session = StatusValidatingSession()
    adapter = create_adapter()
    for schema in SCHEMAS:
        session.mount(schema + "://", adapter)
    session.auth = auth
    session.headers = {
        "Content-Type": "application/json",
        "cache-control": "no-cache",
        "User-Agent": f"up42-py/{version} ({REPOSITORY_URL})",
    }
    return session



"""
Tasking functionality
"""

import dataclasses
from typing import Iterator, List, Literal, Optional, Union

from up42 import base, host, utils

logger = utils.get_logger(__name__)

QuotationDecision = Literal[
    "ACCEPTED",
    "REJECTED",
]
QuotationStatus = Union[Literal["NOT_DECIDED"], QuotationDecision]
FeasibilityStatus = Literal["NOT_DECIDED", "ACCEPTED"]


class QuotationSorting:
    created_at = utils.SortingField(name="createdAt")
    credits_price = utils.SortingField(name="creditsPrice")
    decided_at = utils.SortingField(name="decisionAt")
    updated_at = utils.SortingField(name="updatedAt")


@dataclasses.dataclass
class Quotation:
    session = base.Session()
    id: str
    created_at: str
    updated_at: str
    decided_at: Optional[str]
    account_id: str
    workspace_id: str
    order_id: str
    credits_price: int
    decision: QuotationStatus

    def accept(self):
        self.decision = "ACCEPTED"

    def reject(self):
        self.decision = "REJECTED"

    def save(self):
        url = host.endpoint(f"/v2/tasking/quotation/{self.id}")
        metadata = self.session.patch(url, json={"decision": self.decision}).json()
        quotation = self._from_metadata(metadata)
        for field in dataclasses.fields(quotation):
            setattr(self, field.name, getattr(quotation, field.name))

    @staticmethod
    def _from_metadata(metadata: dict) -> "Quotation":
        return Quotation(
            id=metadata["id"],
            created_at=metadata["createdAt"],
            updated_at=metadata["updatedAt"],
            decided_at=metadata["decisionAt"],
            account_id=metadata["accountId"],
            workspace_id=metadata["workspaceId"],
            order_id=metadata["orderId"],
            credits_price=metadata["creditsPrice"],
            decision=metadata["decision"],
        )

    @classmethod
    def all(
        cls,
        quotation_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        order_id: Optional[str] = None,
        decision: Optional[List[QuotationStatus]] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator["Quotation"]:
        params = {
            "workspaceId": workspace_id,
            "id": quotation_id,
            "orderId": order_id,
            "decision": decision,
            "sort": sort_by,
        }
        return map(
            cls._from_metadata,
            utils.paged_query(params, "/v2/tasking/quotation", cls.session),
        )


class FeasibilityStudySorting:
    created_at = utils.SortingField(name="createdAt")
    updated_at = utils.SortingField(name="updatedAt")
    decided_at = utils.SortingField(name="decisionAt")


@dataclasses.dataclass
class FeasibilityStudyDecisionOption:
    id: str
    description: Optional[str] = None


@dataclasses.dataclass
class FeasibilityStudy:
    session = base.Session()
    id: str
    created_at: str
    updated_at: str
    account_id: str
    workspace_id: str
    order_id: str
    decision: FeasibilityStatus
    options: List[dict]
    decided_at: Optional[str] = None
    decision_option: Optional[FeasibilityStudyDecisionOption] = None

    class NoDecisionOptionChosen(Exception):
        """Raised when trying to save a FeasibilityStudy without a chosen decision option."""

    @classmethod
    def all(
        cls,
        feasibility_study_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        order_id: Optional[str] = None,
        decision: Optional[List[FeasibilityStatus]] = None,
        sort_by: Optional[utils.SortingField] = None,
    ) -> Iterator["FeasibilityStudy"]:
        params = {
            "id": feasibility_study_id,
            "workspaceId": workspace_id,
            "orderId": order_id,
            "decision": decision,
            "sort": sort_by,
        }
        return map(
            cls._from_metadata,
            utils.paged_query(params, "/v2/tasking/feasibility-studies", cls.session),
        )

    @staticmethod
    def _from_metadata(metadata: dict) -> "FeasibilityStudy":
        decision_option = metadata.get("decisionOption")
        if decision_option is not None:
            decision_option = FeasibilityStudyDecisionOption(decision_option["id"], decision_option["description"])
        return FeasibilityStudy(
            id=metadata["id"],
            created_at=metadata["createdAt"],
            updated_at=metadata["updatedAt"],
            account_id=metadata["accountId"],
            workspace_id=metadata["workspaceId"],
            order_id=metadata["orderId"],
            decision=metadata["decision"],
            options=metadata.get("options", []),
            decided_at=metadata.get("decisionAt"),
            decision_option=decision_option,
        )

    def accept(self, option_id: str):
        self.decision_option = FeasibilityStudyDecisionOption(option_id)

    def save(self):
        url = host.endpoint(f"/v2/tasking/feasibility-studies/{self.id}")
        if self.decision_option is None:
            raise FeasibilityStudy.NoDecisionOptionChosen(
                "No decision option chosen for this feasibility study. "
                "Please call 'accept' with a valid option ID before saving."
            )
        metadata = self.session.patch(url, json={"acceptedOptionId": self.decision_option.id}).json()
        feasibility_study = self._from_metadata(metadata)
        for field in dataclasses.fields(feasibility_study):
            setattr(self, field.name, getattr(feasibility_study, field.name))
